package dbtypes

import (
	"errors"
	"fmt"

	"git.andr3h3nriqu3s.com/andr3/fyp/logic/db"
)

type ModelStatus int

const (
	FAILED_TRAINING           ModelStatus = -4
	FAILED_PREPARING_TRAINING             = -3
	FAILED_PREPARING_ZIP_FILE             = -2
	FAILED_PREPARING                      = -1
	PREPARING                             = 1
	CONFIRM_PRE_TRAINING                  = 2
	PREPARING_ZIP_FILE                    = 3
	TRAINING                              = 4
	READY                                 = 5
	READY_ALTERATION                      = 6
	READY_ALTERATION_FAILED               = -6

	READY_RETRAIN        = 7
	READY_RETRAIN_FAILED = -7
)

type ModelDefinitionStatus int

const (
	MODEL_DEFINITION_STATUS_CANCELD_TRAINING ModelDefinitionStatus = -4
	MODEL_DEFINITION_STATUS_FAILED_TRAINING                        = -3
	MODEL_DEFINITION_STATUS_PRE_INIT                               = 1
	MODEL_DEFINITION_STATUS_INIT                                   = 2
	MODEL_DEFINITION_STATUS_TRAINING                               = 3
	MODEL_DEFINITION_STATUS_PAUSED_TRAINING                        = 6
	MODEL_DEFINITION_STATUS_TRANIED                                = 4
	MODEL_DEFINITION_STATUS_READY                                  = 5
)

type ModelHeadStatus int

const (
	MODEL_HEAD_STATUS_PRE_INIT ModelHeadStatus = 1
	MODEL_HEAD_STATUS_INIT                     = 2
	MODEL_HEAD_STATUS_TRAINING                 = 3
	MODEL_HEAD_STATUS_TRAINED                  = 4
	MODEL_HEAD_STATUS_READY                    = 5
)

type BaseModel struct {
	Name         string `json:"name"`
	Status       int    `json:"status"`
	Id           string `json:"id"`
	ModelType    int    `db:"model_type" json:"model_type"`
	ImageModeRaw string `db:"color_mode" json:"image_more_raw"`
	ImageMode    int    `db:"0" json:"image_mode"`
	Width        int    `json:"width"`
	Height       int    `json:"height"`
	Format       string `json:"format"`
	CanTrain     int    `db:"can_train" json:"can_train"`
}

var ModelNotFoundError = errors.New("Model not found error")

func GetBaseModel(db db.Db, id string) (base *BaseModel, err error) {
	var model BaseModel
	err = GetDBOnce(db, &model, "models where id=$1", id)
	if err != nil {
		return
	}
	model.ImageMode = StringToImageMode(model.ImageModeRaw)
	return &model, nil
}

func (m BaseModel) CanEval() bool {
	if m.Status != READY && m.Status != READY_RETRAIN && m.Status != READY_RETRAIN_FAILED && m.Status != READY_ALTERATION && m.Status != READY_ALTERATION_FAILED {
		return false
	}
	return true
}

// DO NOT Pass un filtered data on filters
func (m BaseModel) GetDefinitions(db db.Db, filters string, args ...any) ([]*Definition, error) {
	n_args := []any{m.Id}
	n_args = append(n_args, args...)
	return GetDbMultitple[Definition](db, fmt.Sprintf("model_definition as md where md.model_id=$1 %s", filters), n_args...)
}

func (m BaseModel) GetClasses(db db.Db, filters string, args ...any) ([]*ModelClass, error) {
	n_args := []any{m.Id}
	n_args = append(n_args, args...)
	return GetDbMultitple[ModelClass](db, fmt.Sprintf("model_classes as mc where mc.model_id=$1 %s", filters), n_args...)
}

func (m *BaseModel) UpdateStatus(db db.Db, status ModelStatus) (err error) {
	_, err = db.Exec("update models set status=$1 where id=$2", status, m.Id)
	return
}

type DataPoint struct {
	Id    string `json:"id"`
	Class int    `json:"class"`
	Path  string `json:"path"`
}

func (m BaseModel) DataPoints(db db.Db, mode DATA_POINT_MODE) (data []DataPoint, err error) {
	rows, err := db.Query(
		"select mdp.id, mc.class_order, mdp.file_path from model_data_point as mdp inner "+
			"join model_classes as mc on mc.id = mdp.class_id "+
			"where mc.model_id = $1 and mdp.model_mode=$2;",
		m.Id, mode)
	if err != nil {
		return
	}
	defer rows.Close()

	data = []DataPoint{}

	for rows.Next() {
		var id string
		var class_order int
		var file_path string
		if err = rows.Scan(&id, &class_order, &file_path); err != nil {
			return
		}
		data = append(data, DataPoint{
			Id:    id,
			Path:  file_path,
			Class: class_order,
		})
	}
	return
}

func StringToImageMode(colorMode string) int {
	switch colorMode {
	case "greyscale":
		return 1
	case "rgb":
		return 3
	default:
		panic("unkown color mode")
	}
}
